// SPDX-License-Identifier: BUSL-1.1
pragma solidity 0.8.21;

import "@openzeppelin/contracts/proxy/utils/UUPSUpgradeable.sol";
import "@openzeppelin/contracts/token/ERC721/IERC721.sol";
import "@openzeppelin/contracts/token/ERC1155/IERC1155.sol";
import "@openzeppelin/contracts/token/ERC20/utils/SafeERC20.sol";
import "@openzeppelin/contracts-upgradeable/proxy/utils/Initializable.sol";
import "@openzeppelin/contracts-upgradeable/metatx/ERC2771ContextUpgradeable.sol";
import "@openzeppelin/contracts-upgradeable/token/ERC721/utils/ERC721HolderUpgradeable.sol";
import "@openzeppelin/contracts-upgradeable/token/ERC1155/utils/ERC1155HolderUpgradeable.sol";
import "@openzeppelin/contracts-upgradeable/utils/cryptography/EIP712Upgradeable.sol";
import "../../abstract/roles/EeseeAdmin.sol";
import "../../abstract/BlastPointsReceiverUpgradeable.sol";
import "../../types/Asset.sol";
import "../../interfaces/IDynamicLot.sol";

///@dev Contract for dynamic eesee lots logic.
contract EeseeDynamicLotProvider is 
    UUPSUpgradeable,
    ERC2771ContextUpgradeable, 
    ERC721HolderUpgradeable, 
    ERC1155HolderUpgradeable, 
    EIP712Upgradeable, 
    EeseeAdmin, 
    BlastPointsReceiverUpgradeable 
{
    using SafeERC20 for IERC20;

    struct SignatureData {
        uint256 nonce;
        uint256 deadline;
        bytes signature;
    }

    ///@dev Signer role af defined in {accessManager}.
    bytes32 private constant SIGNER_ROLE = keccak256("SIGNER_ROLE"); 

    ///@dev Lot typehash.
    bytes32 private constant CREATE_LOT_TYPEHASH = keccak256("CreateLot(uint256 ID,bytes32 assetsHash,address owner,address dynamicLot,address sender,bytes32 inputDataHash,uint256 nonce,uint256 deadline)");
    ///@dev Participate typehash.
    bytes32 private constant PARTICIPATE_TYPEHASH = keccak256("Participate(uint256 ID,bytes32 assetsHash,address dynamicLot,address sender,address recipient,bytes32 inputDataHash,uint256 nonce,uint256 deadline)");
    ///@dev Asset transfer typehash.
    bytes32 private constant TRANSFER_ASSET_TYPEHASH = keccak256("TransferAsset(uint256 ID,bytes32 assetsHash,address dynamicLot,address sender,address recipient,bytes32 inputDataHash,bytes32 callHash,uint256 nonce,uint256 deadline)");
    
    ///@dev Is nonce used.
    mapping(uint256 => bool) public nonceUsed;
    ///@dev Is lot exists.
    mapping(uint256 => bool) public lotExists;

    uint256 public constant version = 1;

    event CreateLot(
        uint256 indexed ID, 
        Asset[] assets, 
        address owner,
        address dynamicLot,
        address indexed sender, 
        bytes inputData,
        uint256 indexed nonce
    );
    event Participate(
        uint256 indexed ID, 
        Asset[] assets, 
        address dynamicLot,
        address sender, 
        address indexed recipient, 
        bytes inputData,
        uint256 indexed nonce
    );
    event TransferAsset(
        uint256 indexed ID, 
        Asset[] assets, 
        address dynamicLot,
        address indexed sender, 
        address recipient, 
        bytes inputData,
        uint256 indexed nonce
    );
    event RevokeSignature(uint256 indexed nonce);

    error InvalidRecipient();
    error ExpiredDeadline();
    error NonceUsed();
    error InvalidAmount();
    error InvalidSignature();
    error InvalidValue();
    error TransferNotSuccessful();
    error InvalidAssetType();
    error InvalidLotID();
    
    modifier isValidRecipient(address recipient){
        if(recipient == address(0)) revert InvalidRecipient();
        _;
    }

    constructor(
        IEeseeAccessManager _accessManager,
        address trustedForwarder
    ) ERC2771ContextUpgradeable(trustedForwarder) EeseeRoleHandler(_accessManager) {
        _disableInitializers();
    }

    function initialize(IBlastPoints _blastPoints, address _blastPointsOperator) external initializer {
        __ERC721Holder_init();
        __ERC1155Holder_init();
        __EIP712_init("EeseeDynamicLotProvider", "1");
        __BlastPointsReceiverUpgradeable_init(_blastPoints, _blastPointsOperator);
    }

    // ============ Public Write Functions ============

    /**
     * @dev Creates lot with offchain logic. Emits {CreateLot} event.
     * @param ID - Generated offchain lot ID.
     * @param assets - Assets to list. Note: The sender must have them approved for this contract.
     * @param owner - Lot owner.
     * @param dynamicLot - Dynamic lot address with specific logic for lot.
     * @param inputData - Input data for calling dynamic lot logic.
     * @param signatureData - Signature for asset and data signed by SIGNER_ROLE.
     */
    function createLot(
        uint256 ID,
        Asset[] calldata assets,
        address owner,
        address dynamicLot,
        bytes calldata inputData,
        SignatureData calldata signatureData
    ) external payable {
        address msgSender = _msgSender();
        _checkSignature(
            msgSender, 
            _getCreateLotStructHash(
                ID, 
                assets, 
                owner,
                dynamicLot, 
                msgSender, 
                inputData, 
                signatureData.nonce, 
                signatureData.deadline
            ), 
            signatureData
        );

        if (lotExists[ID]) revert InvalidLotID();
        lotExists[ID] = true;
        IDynamicLot(dynamicLot).createLot(ID, inputData);

        _transferAssetFrom(assets, msgSender);

        emit CreateLot(ID, assets, owner, dynamicLot, msgSender, inputData, signatureData.nonce);
    }

    /**
     * @dev Buys tickets to participate in a lot. Emits {Participate} event.
     * @param ID - ID of lot to buy tickets in.
     * @param assets - Assets to buy tickets with.
     * @param dynamicLot - Dynamic lot address with specific logic for lot.
     * @param recipient - Recipient of tickets.
     * @param inputData - Input data for calling dynamic lot logic.
     * @param signatureData - Signature for ID, asset and recipient signed by SIGNER_ROLE.
     */
    function participate(
        uint256 ID,
        Asset[] calldata assets,
        address dynamicLot,
        address recipient,
        bytes calldata inputData,
        SignatureData calldata signatureData
    ) isValidRecipient(recipient) external payable {
        address msgSender = _msgSender();
        _checkSignature(
            msgSender, 
            _getParticipateStructHash(
                ID, 
                assets, 
                dynamicLot, 
                msgSender, 
                recipient, 
                inputData, 
                signatureData.nonce, 
                signatureData.deadline
            ), 
            signatureData
        );

        if (!lotExists[ID]) revert InvalidLotID();
        IDynamicLot(dynamicLot).participate(ID, inputData);

        _transferAssetFrom(assets, msgSender);

        emit Participate(ID, assets, dynamicLot, msgSender, recipient, inputData, signatureData.nonce);
    }

    /**
     * @dev Transfers asset from a lot. Emits {TransferAsset} event.
     * @param ID - ID of lot to transfer asset in.
     * @param assets - Assets to transfer and send to recipient.
     * @param dynamicLot - Dynamic lot address with specific logic for lot.
     * @param recipient - Recipient of asset.
     * @param inputData - Input data for calling dynamic lot logic.
     * @param _call - Additional external call to make.
     * @param signatureData - Signature for ID, asset, msgSender, recipient and _call signed by SIGNER_ROLE.

     * @return returnData - Data returned from external call.
     */
    function transferAsset(
        uint256 ID,
        Asset[] memory assets, 
        address dynamicLot,
        address recipient,
        bytes calldata inputData,
        bytes memory _call,
        SignatureData calldata signatureData
    ) isValidRecipient(recipient) external payable returns (bytes memory returnData){
        address msgSender = _msgSender();
        _checkSignature(
            msgSender, 
            _getTransferAssetStructHash(
                ID, 
                assets, 
                dynamicLot,
                msgSender, 
                recipient, 
                inputData, 
                _call, 
                signatureData.nonce, 
                signatureData.deadline
            ), 
            signatureData
        );

        if (!lotExists[ID]) revert InvalidLotID();
        IDynamicLot(dynamicLot).transferAsset(ID, inputData);

        _transferAssetTo(assets, recipient);

        if(_call.length != 0) {
            (address target, uint256 value, bytes memory callData) = abi.decode(_call, (address, uint256, bytes));
            if(msg.value != value) revert InvalidValue();
            bool success;
            (success, returnData) = target.call{value: value}(callData);
            if (!success) {
                if (returnData.length == 0) revert();
                assembly { revert(add(32, returnData), mload(returnData)) }
            }
        }
        emit TransferAsset(ID, assets, dynamicLot, msgSender, recipient, inputData, signatureData.nonce);
    }

    /**
     * @dev Callable by SIGNER_ROLE to revoke signatures. Emits {RevokeSignature} for each signature revoked.
     * @param nonces - Signature nonces to revoke.
     */
    function revokeSignatures(uint256[] calldata nonces) external {
        if(!_isSigner(_msgSender())) revert CallerNotAuthorized();
        for(uint256 i; i < nonces.length;){
            if(nonceUsed[nonces[i]] == false){
                nonceUsed[nonces[i]] = true;
                emit RevokeSignature(nonces[i]);
            }
            unchecked { ++i; }
        }
    }

    // ============ Internal Write Functions ============

    function _checkSignature(
        address msgSender,
        bytes32 structHash,
        SignatureData calldata signatureData
    ) internal {
        if(nonceUsed[signatureData.nonce]) revert NonceUsed();
        nonceUsed[signatureData.nonce] = true;
        if(_isSigner(msgSender)) return;
        if(block.timestamp > signatureData.deadline) revert ExpiredDeadline();

        address signer = ECDSAUpgradeable.recover(_hashTypedDataV4(structHash), signatureData.signature);
        if(!_isSigner(signer)) revert InvalidSignature();
    }
    
    function _transferAssetFrom(
        Asset[] memory assets,
        address from
    ) internal {
        for (uint256 i; i < assets.length;) {
            Asset memory asset = assets[i];
            if(asset.amount == 0) {
                if(msg.value != 0) revert InvalidValue();
                continue;
            }
            
            if (asset.assetType == AssetType.Native) {
                if(msg.value != asset.amount) revert InvalidValue();
            } else {
                if(msg.value != 0) revert InvalidValue();
                if (asset.assetType == AssetType.ERC721) {
                    if(asset.amount != 1) revert InvalidAmount();
                    IERC721(asset.token).safeTransferFrom(from, address(this), asset.tokenID, "");
                } else if (asset.assetType == AssetType.ERC1155) {
                    IERC1155(asset.token).safeTransferFrom(from, address(this), asset.tokenID, asset.amount, "");
                } else if (asset.assetType == AssetType.ERC20) {
                    _permit(asset.token, from, asset.data);
                    IERC20(asset.token).safeTransferFrom(from, address(this), asset.amount);
                } else if (asset.assetType == AssetType.ESE) {

                } else revert InvalidAssetType();
            }

            unchecked { ++i; }
        }
    }

    function _transferAssetTo(
        Asset[] memory assets,
        address to
    ) internal {
        for (uint256 i; i < assets.length;) {
            Asset memory asset = assets[i];
            if(asset.amount == 0) continue;
            if (asset.assetType == AssetType.ERC721) {
                IERC721(asset.token).safeTransferFrom(address(this), to, asset.tokenID, asset.data);
            } else if (asset.assetType == AssetType.ERC1155) {
                IERC1155(asset.token).safeTransferFrom(address(this), to, asset.tokenID, asset.amount, asset.data);
            } else if (asset.assetType == AssetType.ERC20) {
                IERC20(asset.token).safeTransfer(to, asset.amount);
            } else if (asset.assetType == AssetType.Native) {
                (bool success,) = to.call{value: asset.amount}(asset.data);
                if(!success) revert TransferNotSuccessful(); 
            } else revert InvalidAssetType();

            unchecked { ++i; }
        }
    }

    function _permit(address token, address spender, bytes memory permit) internal {
        if(permit.length == 0) return;
        (uint256 approveAmount, uint256 deadline, uint8 v, bytes32 r, bytes32 s) = abi.decode(permit, (uint256, uint256, uint8, bytes32, bytes32));
        IERC20Permit(token).permit(spender, address(this), approveAmount, deadline, v, r, s);
    }

    // ============ Internal View Functions ============

    function _getCreateLotStructHash(
        uint256 ID,
        Asset[] calldata assets, 
        address owner,
        address dynamicLot,
        address sender, 
        bytes calldata inputData, 
        uint256 nonce, 
        uint256 deadline
    ) internal pure returns (bytes32){
        return keccak256(abi.encode(
            CREATE_LOT_TYPEHASH, 
            ID,
            keccak256(abi.encode(assets)),
            owner,
            dynamicLot,
            sender,
            keccak256(inputData),
            nonce,
            deadline
        ));
    }

    function _getParticipateStructHash(
        uint256 ID, 
        Asset[] calldata assets, 
        address dynamicLot,
        address sender, 
        address recipient, 
        bytes calldata inputData,
        uint256 nonce, 
        uint256 deadline
    ) internal pure returns (bytes32){
        return keccak256(abi.encode(
            PARTICIPATE_TYPEHASH, 
            ID,
            keccak256(abi.encode(assets)),
            dynamicLot,
            sender,
            recipient,
            keccak256(inputData),
            nonce,
            deadline
        ));
    }

    function _getTransferAssetStructHash(
        uint256 ID, 
        Asset[] memory assets, 
        address dynamicLot,
        address sender, 
        address recipient, 
        bytes calldata inputData,
        bytes memory _call,
        uint256 nonce, 
        uint256 deadline
    ) internal pure returns (bytes32){
        return keccak256(abi.encode(
            TRANSFER_ASSET_TYPEHASH, 
            ID,
            keccak256(abi.encode(assets)),
            dynamicLot,
            sender,
            recipient,
            keccak256(inputData),
            keccak256(_call),
            nonce,
            deadline
        ));
    }

    function _isSigner(address _addr) internal view returns(bool){
        return _hasRole(SIGNER_ROLE, _addr);
    }

    function _msgSender() internal view override(Context, ERC2771ContextUpgradeable) returns (address) {
        return ERC2771ContextUpgradeable._msgSender();
    }

    function _msgData() internal view override(Context, ERC2771ContextUpgradeable) returns (bytes calldata) {
        return ERC2771ContextUpgradeable._msgData();
    }

    function _contextSuffixLength() internal view override(Context, ERC2771ContextUpgradeable) returns (uint256) {
        return ERC2771ContextUpgradeable._contextSuffixLength();
    }

    function _authorizeUpgrade(address newImplementation) internal override onlyAdmin {}
}