// SPDX-License-Identifier: BUSL-1.1
pragma solidity 0.8.21;

import "../../interfaces/IDynamicLot.sol";

contract EeseeLaunchpadLot is IDynamicLot {
    struct Lot {
        uint256 totalTickets;
        uint256 boughtTickets;
    }

    struct CreateLotInput {
        uint256 totalTickets;
    }

    struct ParticipateInput {
        uint256 amount;
        address recipient;
        uint256 maxTicketsForUser;
    }

    ///@dev Dynamic lot provider contract address.
    address public dynamicLotProvider;

    ///@dev Mapping stores launchpad lots by lot id.
    mapping(uint256 => Lot) public lots;
    ///@dev Mapping stores user purchases in each lot.
    mapping(uint256 => mapping(address => uint256)) public participants;

    error TotalTicketsOverflow();
    error MaxTicketsForUserOverflow();
    error InvalidCaller();

    modifier onlyDynamicLotProvider() {
        if (msg.sender != dynamicLotProvider) revert InvalidCaller();
        _;
    }

    constructor(address _dynamicLotProvider) {
        dynamicLotProvider = _dynamicLotProvider;
    }

    function createLot(uint256 ID, bytes calldata inputData) external onlyDynamicLotProvider {
        (CreateLotInput memory input) = abi.decode(inputData, (CreateLotInput));

        lots[ID] = Lot({
            totalTickets: input.totalTickets,
            boughtTickets: 0
        });
    }

    function participate(uint256 ID, bytes calldata inputData) external onlyDynamicLotProvider {
        (ParticipateInput memory input) = abi.decode(inputData, (ParticipateInput));

        Lot memory lot = lots[ID];
        if (lot.boughtTickets + input.amount > lot.totalTickets) revert TotalTicketsOverflow();
        if (input.amount + participants[ID][input.recipient] > input.maxTicketsForUser) revert MaxTicketsForUserOverflow();

        lots[ID].boughtTickets += input.amount;
        participants[ID][input.recipient] += input.amount;
    }

    function transferAsset(uint256 /*ID*/, bytes calldata /*inputData*/) external onlyDynamicLotProvider {}
}