// SPDX-License-Identifier: BUSL-1.1
pragma solidity 0.8.21;

import "@openzeppelin/contracts/token/ERC1155/ERC1155.sol";
import "@openzeppelin/contracts/utils/cryptography/ECDSA.sol";
import "@openzeppelin/contracts/utils/cryptography/EIP712.sol";
import "@openzeppelin/contracts/metatx/ERC2771Context.sol";
import "../abstract/roles/EeseeAdmin.sol";

contract EeseeRWA is ERC1155, EeseeAdmin, EIP712, ERC2771Context {
    struct SignatureData{
        uint256 nonce;
        uint256 deadline;
        bytes signature;
    }

    ///@dev Signer role af defined in {accessManager}.
    bytes32 private constant SIGNER_ROLE = keccak256("SIGNER_ROLE"); 

    bytes32 private constant MINT_TYPEHASH = keccak256("Mint(address to,uint256 id,uint256 amount,uint256 nonce,uint256 deadline)");
    bytes32 private constant LOCK_TYPEHASH = keccak256("Lock(address account,uint256 id,uint256 amount,bytes encryptedData,uint256 nonce,uint256 deadline)");
    bytes32 private constant UNLOCK_TYPEHASH = keccak256("Unlock(address account,uint256 id,uint256 amount,uint256 nonce,uint256 deadline)");

    mapping(uint256 => bool) public nonceUsed;
    mapping(uint256 => mapping(address => uint256)) private _lockedBalances;

    event RevokeSignature(uint256 indexed nonce);
    event Locked(address indexed account, uint256 indexed id, uint256 amount, bytes encryptedData);
    event Unlocked(address indexed account, uint256 indexed id, uint256 amount);

    error ExpiredDeadline();
    error NonceUsed();
    error InvalidSignatureDataLength();
    error InvalidAmount();
    error InvalidSignature();

    constructor(
        string memory _uri, 
        IEeseeAccessManager _accessManager,
        address trustedForwarder
    ) ERC1155(_uri) EeseeRoleHandler(_accessManager) EIP712("EeseeRWA", "1") ERC2771Context(trustedForwarder){}

    // ============ Public View Functions ============

    function balanceOf(address account, uint256 id) public view override returns (uint256) {
        return super.balanceOf(account, id) - _lockedBalances[id][account];
    }

    // ============ External Write Functions ============

    /**
     * @dev Callable by SIGNER_ROLE or by accounts with signature, who want to mint tokens.
     * @param to - Account address to mint tokens to.
     * @param id - Token id to mint.
     * @param amount - Token amount to mint.
     * @param data - Signed message, if minter is not signer.
     */
    function mint(address to, uint256 id, uint256 amount, bytes memory data) external {
        _mint(to, id, amount, data);
    }

    /**
     * @dev Callable by SIGNER_ROLE or by holders, who want to lock their tokens.
     * @param account - Account address to lock.
     * @param id - Token id to lock.
     * @param amount - Token amount to lock.
     * @param encryptedData - Encrypted delivery data.
     * @param signatureData - Signed message with nonce and deadline.
     */
    function lock(address account, uint256 id, uint256 amount, bytes calldata encryptedData, SignatureData calldata signatureData) external {
        if (balanceOf(account, id) < amount) revert InvalidAmount();
        _checkSignature(
            _msgSender(),
            keccak256(abi.encode(LOCK_TYPEHASH, account, id, amount, encryptedData, signatureData.nonce, signatureData.deadline)),
            signatureData
        );
        
        _lockedBalances[id][account] += amount;
        emit Locked(account, id, amount, encryptedData);
    }

    /**
     * @dev Callable by SIGNER_ROLE or by holders, who wants to unlock their tokens.
     * @param account - Account address to unlock.
     * @param id - Token id to unlock.
     * @param amount - Token amount to unlock.
     * @param signatureData - Signed message with nonce and deadline.
     */
    function unlock(address account, uint256 id, uint256 amount, SignatureData calldata signatureData) external {
        uint256 lockedBalance = _lockedBalances[id][account];
        if (lockedBalance < amount) revert InvalidAmount();
        _checkSignature(
            _msgSender(), 
            keccak256(abi.encode(UNLOCK_TYPEHASH, account, id, amount, signatureData.nonce, signatureData.deadline)),
            signatureData
        );

        _lockedBalances[id][account] = lockedBalance - amount;
        emit Unlocked(account, id, amount);
    }

    /**
     * @dev Callable by SIGNER_ROLE to revoke signatures. Emits {RevokeSignature} for each signature revoked.
     * @param nonces - Signature nonces to revoke.
     */
    function revokeSignatures(uint256[] calldata nonces) external {
        if(!_isSigner(_msgSender())) revert CallerNotAuthorized();
        for(uint256 i; i < nonces.length;){
            if (nonceUsed[nonces[i]] == false){
                nonceUsed[nonces[i]] = true;
                emit RevokeSignature(nonces[i]);
            }
            unchecked { ++i; }
        }
    }

    // ============ Internal Write Functions ============

    function _beforeTokenTransfer(
        address operator, 
        address from, 
        address to, 
        uint256[] memory ids, 
        uint256[] memory amounts, 
        bytes memory data
    ) internal override {
        if (from == address(0)) { // mint
            (SignatureData[] memory signaturesData, bytes memory additionalData) = abi.decode(data, (SignatureData[], bytes));
            if (signaturesData.length != ids.length) revert InvalidSignatureDataLength();

            for (uint256 i; i < ids.length;) {
                _checkSignature(
                    operator, 
                    keccak256(abi.encode(MINT_TYPEHASH, to, ids[i], amounts[i], signaturesData[i].nonce, signaturesData[i].deadline)),
                    signaturesData[i]
                );
                unchecked { ++i; }
            }
            data = additionalData;
        } else {
            for (uint256 i; i < ids.length;) {
                if (amounts[i] > balanceOf(from, ids[i])) revert InvalidAmount();
                unchecked { ++i; }
            }
        }
    }

    function _checkSignature(
        address msgSender,
        bytes32 structHash,
        SignatureData memory signatureData
    ) internal {
        if(nonceUsed[signatureData.nonce]) revert NonceUsed();
        nonceUsed[signatureData.nonce] = true;
        if(_isSigner(msgSender)) return;
        if(block.timestamp > signatureData.deadline) revert ExpiredDeadline();

        address signer = ECDSA.recover(_hashTypedDataV4(structHash), signatureData.signature);
        if(!_isSigner(signer)) revert InvalidSignature();
    }

    // ============ Internal View Functions ============

    function _isSigner(address _addr) internal view returns(bool){
        return _hasRole(SIGNER_ROLE, _addr);
    }

    function _msgSender() internal view override(Context, ERC2771Context) returns (address) {
        return ERC2771Context._msgSender();
    }

    function _msgData() internal view override(Context, ERC2771Context) returns (bytes calldata) {
        return ERC2771Context._msgData();
    }

    function _contextSuffixLength() internal view override(Context, ERC2771Context) returns (uint256) {
        return ERC2771Context._contextSuffixLength();
    }
}
